<copyright> TOM parser generator bootstrap parser.
    Written by <a href="mailto:tiggr@ics.ele.tue.nl">Pieter J. Schoenmakers</a>

    Copyright (C) 1996 Pieter J. Schoenmakers.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
    MA 02111-1307, USA

    <id>$Id: GPParse.t,v 1.39 1998/11/10 17:28:42 tiggr Exp $</id>
    </copyright>

/******************** GPParseDefines ********************/

implementation class
GPParseDefines: ParserDefines
{
  <doc> The string to output for a new line.  </doc>
  static MutableByteString indent;

  <doc> Repitition of elements of a rule.  </doc>
  const REPEAT_OPTIONAL = -1;
  const REPEAT_STAR = -2;
  const REPEAT_PLUS = -3;

  <doc> The token value (to be used in the look ahead set) for the empty
      production rule (i.e. epsilon).  </doc>
  const EPSILON = 65536;

  const IDENTIFIER = 65537;

  const NUMBER = 65538;

  /* Something between `{'... `}'.  */
  const EXPR = 65539;

  /* Something between `'' characters.  */
  const LITERAL = 65540;

  /* Real tokens.  */
  const TOKEN = 65541;
  const CLASS = 65542;
  const INSTANCE = 65543;

  /* Something between %% and %%, taken as literal method definitions.  */
  const CODE = 65544;

  /* A token, the value of which is not defined by GP.  */
  const PREDEF = 65545;

  <doc> Iff YES, there were changes in the sets.  </doc>
  static boolean changes;
}

end;

implementation instance GPParseDefines end;

/******************** GPParse ********************/

implementation class
GPParse: Parser, GPParseDefines, gp
{
  <doc> Iff {TRUE} emit cost table.  </doc>
  mutable static boolean verbose_cost;

  <doc> Iff {TRUE} emit flat rules.  </doc>
  mutable static boolean verbose_flat_rules;
}

void
  fill_tokens_by_number
{
}

end;

implementation instance
GPParse
{
  <doc> The things declared to be a token, keyed on their name.  </doc>
  MutableDictionary tokens;

  <doc> The non-terminals, keyed on their name.  </doc>
  MutableDictionary non_terminals;

  <doc> All elements from Vt in a closed naming.  </doc>
  public MutableArray v-terminals;

  <doc> Mapping from token number to token name for the parser we are
      creating.  </doc>
  public MutableArray gp_tokens_by_number;

  // It isn't really public; it is available to our `friend' NonTerminal...
  // Wed Jun 12 15:34:39 1996, tiggr@cobra.es.ele.tue.nl
  <doc> The rules, keyed on the name of the non-terminal.  </doc>
  public MutableDictionary rules;

  <doc> The superclasses of the parser class.  </doc>
  MutableIndexed superclasses;

  <doc> The expressions outside rules before the first rules: class
      methods.  </doc>
  MutableArray class_methods;

  <doc> The expressions outside rules after the first rules: instance
      methods.  </doc>
  MutableArray instance_methods;

  <doc> The instance variables expression.  </doc>
  String ivars;

  <doc> The name of the parser class.  </doc>
  String parser_name;

  <doc> The top rule.  </doc>
  Rule top;

  <doc> The next numeric value to be assigned to a token.  </doc>
  int next_token;
}

Token (t)
  addToken String ident
{
  if (tokens[ident] != nil)
    [[[self error] print ("redeclaration of token `", ident, '\'')] nl];
  else
    {
      t = [[Token alloc] init ident value next_token++];
      [gp_tokens_by_number add ident];
      tokens[ident] = t;

      [self addTerminal t];
    }
}

<doc> Return the {Terminal} with the indicated token value.  Create it if
    needed.  </doc>
Terminal
  terminal int terminal_value
pre
  terminal_value >= 0 && terminal_value < 65536
{
  int i, n = [v-terminals length];
  for (i = 0; i < n; i++)
    {
      Terminal t = v-terminals[i];
      if ([t token_value] == terminal_value)
	return t;
    }

  Terminal t = [[Terminal alloc] init terminal_value];
  [self addTerminal t];
  = t;
}

<doc> Add the (newly created) terminal (or token) {t} to the {v-terminals},
    setting its rank.  </doc>
protected void
  addTerminal Terminal t
pre
  t != nil
{
  [t set_rank [v-terminals length]];
  [v-terminals add t];

  /* The default cost of terminals and tokens is 1.  */
  [t set_cost 1];
}

<doc> Describe the {token} of the generated parser on the output stream
    {s}.  </doc>
OutputStream (s)
  describeGenerated OutputStream s
	      token int token
{
  if (token >= ' ' && token <= '~')
    [s print ('\'', byte (token), '\'')];
  else if (token == 0)
    [s print "EOF"];
  else if (token >= 65536)
    [s print ({Any t = gp_tokens_by_number[token - 65536];})];
  else
    [s print ("char(", token, ')')];
}

<doc> Describe the tokens of the generated parser from the enumerator {e}
    on the output stream {s}.  </doc>
OutputStream
  describeGenerated OutputStream s
	     tokens Enumerator e
{
  boolean first = YES;
  int i;

  // GGG See comment with [Element condition]
  // Sun Dec 29 15:35:44 1996, tiggr@tricky.es.ele.tue.nl
  while ({boolean b; (b, i) = [e next]; b == TRUE;})
    {
      if (first)
	first = NO;
      else
	s = [s print ' '];
      s = [self describeGenerated s token i];
    }
  = s;
}

void
  dumpSets OutputStream str
{
  Enumerator e;
  boolean b;
  Rule r;

  e = [rules enumerator];
  while ({(b, r) = [e next]; b;})
    {
      [str print ([r name], " first: ")];
      if (!![r first])
	[self describeGenerated str tokens [[r first] enumerator]];
      else
	[str print nil];

      [[str nl] print ([r name], " la: ")];
      if (!![r la])
	[self describeGenerated str tokens [[r la] enumerator]];
      else
	[str print nil];

      [[str nl] print ([r name], " follow: ")];
      if (!![r follow])
	[self describeGenerated str tokens [[r follow] enumerator]];
      else
	[str print nil];

      [str nl];
    }
}

OutputStream
  generateDefines OutputStream str
{
  Enumerator e;
  boolean b;
  String s;
  Token t;

  [str print ("implementation class ", parser_name,
	      "Defines: gps.ParserDefines\n{")];

  e = [tokens keyEnumerator];
  while ({(b, s) = [e next]; b;})
    {
      Token t = tokens[s];
      if (![t predef])
	[[str print (indent, "const ", s, " = ", [tokens[s] token_value], ';')]
	  nl];
    }

  [[str print (indent, '}', indent, "end;")] nl];

  [[str print ("implementation instance ", parser_name, "Defines end;")] nl];
}

OutputStream
     generate OutputStream str
  definitions OutputStream def_str
{
  Enumerator e;
  boolean b;
  String s;
  Rule r;

  current_parser = self;

  /* Walk the non-terminals, checking they have been defined.  */
  e = [non_terminals keyEnumerator];
  while ({(b, s) = [e next]; b;})
    if (!rules[s])
      [[[self error] print ('`', s, "' used but not defined")] nl];

  if (!top)
    [[[self error] print "no top rule"] nl];
  else if (!parser_name)
    [[[self error] print "no parser class name"] nl];
  else if (!num_errors)
    {
      IntegerRangeSet eof = [IntegerRangeSet new];

      [eof add EOF];

      [top set_is_top YES];

      /* Make the flatrules.  */ 
      Rule r;
      e = [rules valueEnumerator];
      while ({(b, r) = [e next]; b;})
	[r makeFlatRules];

      /* Scrutinize the rules, checking for rules which are never used.  */
      e = [rules keyEnumerator];
      while ({(b, s) = [e next]; b;})
	{
	  NonTerminal nt = non_terminals[s];
	  if (![nt used] && top != rules[s])
	    [[err print ("warning: rule `", s, "' defined but not used")] nl];
	}

      if (verbose_flat_rules)
	{
	  [[[stdio err] print "Production rules:"] nl];
	  [non_terminals makeElementsPerform
			 selector (void describe-flat-rules)];
	  [[stdio err] nl];
	}

      do
	{
	  changes = NO;
	  [Element next_run];
	  [top computeLA eof];
	} while (changes);

      if (verbose)
	[self dumpSets err];

      /* Error recovery preparation. Yum */

      /* Initialize terminal insertion costs and S-elements Yum*/
      // TODO: Terminal insertion cost vector can be supplied by input. Yum
      [self computeSTable];
      [self computeETable];

      if (!num_errors)
	{
	  [top checkLL1];

	  MutableByteString header
	    = [[MutableByteString new]
	       print ("<doc> Generated by ", [Runtime program_name],
		      ".  </doc>\n")];

	  [header write str];
	  if (def_str != nil)
	    [header write def_str];

	  [self generateDefines !def_str ? str : def_str];

	  [str print ("implementation class ", parser_name,
		      ": gps.Parser, ", parser_name, "Defines")];
	  if (superclasses != nil)
	    {
	      int i, n = [superclasses length];
	      for (i = 0; i < n; i++)
		{
		  String name = superclasses[i];
		  [str print (", ", name)];
		}
	    }
	  [str nl];

	  [str print "void\n  fill_tokens_by_number\n{
tokens_by_number = [MutableObjectArray new];\n"];

	  [str print ("[tokens_by_number resize ", 1 + [tokens length],
		      "];\n")];
	  e = [tokens keyEnumerator];
	  {
	    String key;

	    while ({(b, key) = [e next]; b;})
	      {
		Token t = tokens[key];
		// What to do with predef's?
		// Fri Apr 10 20:26:04 1998, tiggr@gerbil.org
		[str print ("tokens_by_number[", [t token_value] - 65536,
			    "] = \"", key, "\";\n")];
	      }
	  }
	  
	  [str print "}\n"];

	  if (class_methods != nil)
	    {
	      int i, n = [class_methods length];

	      for (i = 0; i < n; i++)
		[[str print ("\n", ({All a = class_methods[i];}))] nl];
	    }

	  [[str print "end;"] nl];
	  [[str print ("implementation instance ", parser_name)] nl];

	  if (ivars != nil)
	    [[str print ("{\n", ivars, "\n}")] nl];

	  if (instance_methods != nil)
	    {
	      int i, n = [instance_methods length];

	      for (i = 0; i < n; i++)
		[[str print ("\n", {All a = instance_methods[i];})] nl];
	    }

	  e = [rules enumerator];
	  while ({(b, r) = [e next]; b;})
	    [r compile str];
	  [str nl];

	  [[str print ("end;")] nl];
	}
    }

  current_parser = nil;

  = str;
}

<doc> Override designated initializer.  </doc>
id
  init Lexer l
{
  rules = [MutableDictionary new];
  tokens = [MutableDictionary new];
  non_terminals = [MutableDictionary new];
  v-terminals = [MutableObjectArray new];

  next_token = 65537;

  = [super init l];
}

<doc> Return the non-terminal with the {name}.  May only be used for
    existing non-terminals.  </doc>
NonTerminal (result)
  nonTerminal String name
post
  result != nil
{
  = non_terminals[name];
}

void
  addNonTerminal NonTerminal nt
pre
  !non_terminals[[nt name]]
{
  non_terminals[[nt name]] = nt;
}

/******************** grammar rules ********************/

<doc> This is the top rule.  </doc>
void
  grammar
{
  if (trace)
    [[err print ("entering grammar")] nl];

  indent = [MutableByteString new];
  [indent add '\n'];

  current_parser = self;

  [self nextToken];

  /* Initialize.  */
  gp_tokens_by_number = [MutableObjectArray new];
  [gp_tokens_by_number add "EPSILON"];

  while (token != EOF)
    {
      if (token == IDENTIFIER || token == TOKEN || token == PREDEF
	  || token == CLASS || token == CODE || token == INSTANCE)
	[self declaration];
      else
	[self fatalError];
    }

  current_parser = nil;
}

private void
  declaration
{
  if (trace)
    [[err print ("entering declaration")] nl];

  if (token == IDENTIFIER)
    [self rule];
  else if (token == TOKEN)
    {
      [self nextToken];

      for (;;)
	{
	  if (token == IDENTIFIER)
	    {
	      [self addToken [lex semantics]];
	      [self nextToken];
	    }
	  else
	    [self fatalError];

	  if (token == ',')
	    [self nextToken];
	  else
	    break;
	}
    }
  else if (token == PREDEF)
    for (;;)
      {
	[self nextToken];
	if (token != IDENTIFIER)
	  [self fatalError];

	Token t = [self addToken [lex semantics]];
	[t set_predef YES];
	[self nextToken];

	if (token != ',')
	  break;
      }
  else if (token == CLASS)
    {
      [self nextToken];

      if (token == IDENTIFIER)
	{
	  if (!parser_name)
	    parser_name = [lex semantics];
	  else
	    [[[self error] print "parser class name already defined"] nl];
	  [self nextToken];

	  if (token == ':')
	    {
	      [self nextToken];

	      superclasses = [MutableObjectArray new];
	      while (TRUE)
		{
		  String name = [lex semantics];

		  [self term IDENTIFIER];

		  if (token == '.')
		    {
		      [self nextToken];
		      name = [[MutableByteString new]
			       print (name, '.', [lex semantics])];
		      [self term IDENTIFIER];
		    }

		  [superclasses add name];

		  if (token == ',')
		    [self nextToken];
		  else
		    break;
		}
	    }
	}
      else
	[self fatalError];
    }
  else if (token == INSTANCE)
    {
      [self nextToken];
      if (token == EXPR)
	{
	  if (!ivars)
	    ivars = [lex semantics];
	  else
	    [[[self error] print "instance variables already defined"] nl];
	  [self term EXPR];
	}
      instance_methods = [MutableObjectArray new];
    }
  else if (token == CODE)
    {
      if (!instance_methods && !ivars && ![rules length])
	{
	  if (!class_methods)
	    class_methods = [MutableObjectArray new];
	  [class_methods add [lex semantics]];
	}
      else
	[instance_methods add [lex semantics]];

      [self nextToken];
    }
  else
    [self fatalError];
}

private void
  rule
{
  String expr;

  if (trace)
    [[err print ("entering rule")] nl];

  if (token != IDENTIFIER)
    [self fatalError];

  String nt = [lex semantics];
  if (tokens[nt] != nil)
    [[[self error] print ("token ", nt, " used as non-terminal")] nl];
  else if (rules[nt] != nil)
    [[[self error] print ("multiple rules for `", nt, '\'')] nl];
  else if (!non_terminals[nt])
    non_terminals[nt] = [[NonTerminal alloc] init nt];

  [self nextToken];

  Array attributes;

  if (token == '(')
    attributes = [self attributes];

  if (token == EXPR)
    {
      expr = [lex semantics];
      [self nextToken];
    }

  [self term ':'];

  Rule r = [[Rule alloc] init nt start expr productions [self alternatives]
			 attributes attributes];
  if (top == nil)
    top = r;
  rules[nt] = r;

  if (token == ':')
    {
      [self nextToken];
      [r set_post_expr [lex semantics]];
      [self term EXPR];
    }

  [self term ';'];
}

<doc> Grammar rule to parse all branches of a {[ ... | ... | ... ]}.
    </doc>
private Alternatives
  alternatives
{
  Alternatives r = [Alternatives new];

  if (trace)
    [[err print ("entering alternatives")] nl];

  [r add [self alternative]];

  while (token == '|')
    {
      [self nextToken];
      [r add [self alternative]];
    }

  = r;
}

<doc> Grammar rule to parse one branch of a {[ ... | ... | ... ]}.
    </doc>
private Sequence
  alternative
{
  Sequence r = [Sequence new];

  if (trace)
    [[err print ("entering alternative")] nl];

  while (token != ';' && token != ']' && token != '|' && token != ':')
    if (token == EXPR)
      {
	[r add [[Annotation alloc] init nil attributes nil
				   expr [lex semantics]]];
	[self nextToken];
      }
    else
      [r add [self repeats [self element]]];

  = r;
}

private Annotation
  element
{
  String expr;
  Array a;

  if (trace)
    [[err print ("entering element")] nl];

  if (token == '[')
    {
      [self nextToken];

      = [[Annotation alloc] init [self alternatives] attributes nil expr nil];

      [self term ']'];
    }
  else if (token == IDENTIFIER)
    {
      String name = [lex semantics];
      Symbol s = tokens[name];

      if (!s)
	{
	  s = non_terminals[name];
	  if (!s)
	    non_terminals[name]
	      = s = [[NonTerminal alloc] init name];
	}

      [self nextToken];
      if (token == '(')
	a = [self actualAttributes];

      if (token == EXPR)
	{
	  expr = [lex semantics];
	  [self nextToken];
	}

      = [[Annotation alloc] init s attributes a expr expr];
    }
  else if (token == LITERAL)
    {
      Terminal lit = [lex semantics];

      [self nextToken];

      if (token == '(')
	{
	  a = [self actualAttributes];
	  String a_name = !a || [a length] != 1 ? nil : a[0];

	  if (!a_name)
	    [[[Element name [[self error] print "bad attributes to token `"]
		       token [lit token_value]] print '\''] nl];
	}

      if (token == EXPR)
	{
	  expr = [lex semantics];
	  [self nextToken];
	}

      = [[Annotation alloc] init lit attributes a expr expr];
    }
  else
    [self fatalError];
}

private Annotation
  repeats Annotation e
{
  if (trace)
    [[err print ("entering repeats")] nl];

  if (token == '*')
    {
      [e set_repeats REPEAT_STAR];
      [self nextToken];
    }
  else if (token == '+')
    {
      [e set_repeats REPEAT_PLUS];
      [self nextToken];
    }
  else if (token == '?')
    {
      [e set_repeats REPEAT_OPTIONAL];
      [self nextToken];
    }

  = e;
}

private Array
  actualAttributes
{
  MutableArray attr = [MutableObjectArray new];

  if (trace)
    [[err print ("entering attributes")] nl];

  do
    {
      [self nextToken];

      if (token == EXPR || token == IDENTIFIER)
	{
	  [attr add [lex semantics]];
	  [self nextToken];
	}
      else
	[self fatalError];
    } while (token == ',');

  [self term ')'];

  = attr;
}

private Array
  attributes
{
  MutableArray attr = [MutableObjectArray new];

  if (trace)
    [[err print ("entering attributes")] nl];

  [self term '('];
  [attr add [self attribute]];

  do
    if (token == ',')
      {
	[self nextToken];
	[attr add [self attribute]];
      }
    while (token != ')');

  [self nextToken];

  = attr;
}

private Attribute
  attribute
{
  if (trace)
    [[err print ("entering attribute")] nl];

  if (token == '>' || token == '<')
    {
      boolean in_p = token == '>';
      String type, name;

      [self nextToken];
      type = [lex semantics];
      [self term IDENTIFIER];

      if (token == '(')
	{
	  [self nextToken];

	  String real_type = [lex semantics];
	  [self term IDENTIFIER];
	  [self term ')'];

	  type = [[MutableByteString new] print (type, '(', real_type, ')')];
	  [type freeze];
	}

      name = [lex semantics];
      [self term IDENTIFIER];

      = [[Attribute alloc] init (type, name) in in_p];
    }
  else
    [self fatalError];
}

OutputStream (s)
  costTable OutputStream s
{
  int i, n = [v-terminals length];
  for (i = 0; i < n; i++)
    {
      Terminal t = v-terminals[i];
      [[[t describe-flat-rule s] print (":\t", [t cost])] nl];
    }

  Enumerator e = [non_terminals enumerator];
  NonTerminal nt;
  boolean b;
  while ({(b, nt) = [e next]; b;})
    [[s print ([nt name], ":\t", [nt cost])] nl];
}

<doc> Iterate over all non terminals A and determine C(A) and S(A).
    Yum.  </doc>
private void
  computeSTable
{
  if (trace)
    [[err print "Computation of S-Table started"] nl];

  boolean b, changes;
  NonTerminal nt;
  Enumerator e;

  do
    {
      boolean verbose_progress = verbose_cost && verbose_flat_rules;
      changes = FALSE;

      e = [non_terminals enumerator];
      while ({(b, nt) = [e next]; b;})
	{
	  Array flat_rules = [nt flat_rules];
	  int i, n = [flat_rules length];

	  for (i = 0; i < n; i++)
	    {
	      FlatRule f_r = flat_rules[i];
	      int cost_rhs = [f_r cost];
	      int cost_lhs = [nt cost];

	      if (cost_rhs != -1 && (cost_lhs == -1 || cost_rhs < cost_lhs))
		{
		  /* Costs have changed for this flat rule.  */
		  if (verbose_progress)
		    {
		      [[stdio err] print (cost_lhs, " -> ", cost_rhs, "\t")]; 
		      [f_r describe-flat-rule [stdio err]];
		    }

		  [nt set_cost cost_rhs];
		  changes = TRUE;

		  //		  [[nt S-element] empty];
		  //[[nt S-element] add [f_r rhs]];
		}
	    }
	}

      if (changes && verbose_progress)
	[[[stdio err] print '.'] nl];
    } while (changes);

  if (verbose_cost)
    [self costTable [stdio err]];

  if (verbose)
    [self describeSTable];
}

void
  describeSTable
{
  NonTerminal nt;
  Enumerator e;
  boolean b;
 
  int i, n = [v-terminals length];
  for (i = 0; i < n; i++)
    {
      Terminal t = v-terminals[i];
      //      [[[stdio err] print ("Terminal ", t, "\t ", [t S-element])] nl];
    }
  
  e = [non_terminals enumerator];
  while ({(b, nt) = [e next]; b;})
    {
      //      [[[stdio err] print ("Non-terminal: ", nt, "\t", [nt S-element])] nl];
    }
}

private void
  computeETable
{
  if (trace)
    [[err print "Computation of E-Table started"] nl];
}

end;
